import { ViewChild, Input, Output, OnDestroy, EventEmitter, Directive } from "@angular/core";
import { Subscription } from "rxjs";
import { Table } from "primeng/table";
import * as moment from "moment";
import { BaseCRUDServiceInterface } from "../shared/base-interface.service";
import { Domain } from "../../shared/data/domain";
import { Item } from "../../shared/data/item";
import { DomainCachedService } from "../../shared/service/domain-cached.service";
import { TableViewService } from "../../shared/service/tableview.service";
import { BaseFilter } from "../shared/base.filter";

/**
 * Clase base de los componentes TableView
 */
@Directive()
export class BaseTableViewComponent implements OnDestroy {
  @ViewChild("dataTable") public dataTable: Table;

  items: Item[];

  @Input() totalItems: number = 0;
  @Output() private onPageChange = new EventEmitter<Item>();

  protected dataService: BaseCRUDServiceInterface;
  protected domainCachedService: DomainCachedService;
  protected tableViewService: TableViewService;

  csvSeparator: string = ",";

  private subscriptionMessages: Subscription;
  protected RESULTADOS_PAGE = BaseFilter.RESULTADOS_PAGE; // se hace visible esta variable a cualquier subclase de tableview que pudiera necesitarla

  constructor(domainCachedService: DomainCachedService, tableViewService: TableViewService) {
    this.domainCachedService = domainCachedService;
    this.tableViewService = tableViewService;
    this.subscriptionMessages = tableViewService.observableExport.subscribe((message) => {
      this.exportCSV();
    });
  }

  exportCSV() {
    let dt = this.dataTable;
    if (dt == null) {
      alert("Exportació no disponible");
      return;
    }
    dt.exportCSV();
  }

  getDataTable(): Table {
    return null;
  }

  getColumna(record: Item, field: string) {
    return null;
  }

  getColumnaDominio(dominio: Domain[], valor: Object) {
    let domainMatch =
      dominio &&
      dominio.find((obj) => {
        return obj.value == valor;
      });
    if (domainMatch) {
      return domainMatch.label;
    } else if (valor) {
      return String(`??${valor}??`);
    }
    return "";
  }

  ngOnDestroy() {
    this.subscriptionMessages.unsubscribe();
  }

  pageChanged(event): void {
    console.log("pageChanged(" + event + ")");
    this.onPageChange.emit(event.page + 1);
  }

  /**
   * Método de ordenación de campos de fecha.
   * @see Adaptación de: https://stackoverflow.com/questions/41181277/primeng-datatable-custom-sorting-or-filtering-angular-2
   */
  dateSort(event, field) {
    const comparer = function (a, b): number {
      const formatedA = a[field] ? moment(a[field], "DD/MM/YYYY").format("YYYY-MM-DD") : "-";
      const formatedB = b[field] ? moment(b[field], "DD/MM/YYYY").format("YYYY-MM-DD") : "-";
      // let result = -1; if (moment(formatedB).isBefore(formatedA, 'day')) { result = 1; }
      const result = formatedA.localeCompare(formatedB);
      return result * event.order;
    };

    const strcmp = function (a, b): number {
      if (a.toString() < b.toString()) {
        return -1;
      }
      if (a.toString() > b.toString()) {
        return 1;
      }
      return 0;
    };

    this.items.sort(comparer);
    this.items = [...this.items];
  }

  /**Función de ordenación de columnas de los p-table de prime */
  customColumnsSort(event: any, items: Item[]): Item[] {
    let field: string = event.field;
    let vregexFecha = /^(0[1-9]|[1-2]\d|3[01])(\/)(0[1-9]|1[012])\2(\d{4})$/;
    let vregexFechaHora = /^(0[1-9]|[1-2]\d|3[01])(\/)(0[1-9]|1[012])\2(\d{4}) [012][0-9]:[0-6][0-9]$/;
    items.sort((a, b) => {
      // comparamos DD/MM/YYYY
      if (vregexFecha.test(a[field]) && vregexFecha.test(b[field])) {
        let fechaA = a[field].substring(6) + a[field].substring(3, 5) + a[field].substring(0, 2);
        let fechaB = b[field].substring(6) + b[field].substring(3, 5) + b[field].substring(0, 2);
        const sortDesc = fechaA < fechaB ? (event.order === 1 ? 1 : -1) : 0;
        return fechaA > fechaB ? (event.order === 1 ? -1 : 1) : sortDesc;
      }
      // comparamos DD/MM/YYYY HH:MM
      else if (vregexFechaHora.test(a[field]) && vregexFechaHora.test(b[field])) {
        let fechaHoraA =
          a[field].substring(6, 10) +
          a[field].substring(3, 5) +
          a[field].substring(0, 2) +
          a[field].substring(11, 13) +
          a[field].substring(14);
        let fechaHoraB =
          b[field].substring(6, 10) +
          b[field].substring(3, 5) +
          b[field].substring(0, 2) +
          b[field].substring(11, 13) +
          b[field].substring(14);
        const sortDesc = fechaHoraA < fechaHoraB ? (event.order === 1 ? 1 : -1) : 0;
        return fechaHoraA > fechaHoraB ? (event.order === 1 ? -1 : 1) : sortDesc;
      } else {
        const sortDesc = a[field] < b[field] ? (event.order === 1 ? 1 : -1) : 0;
        return a[field] > b[field] ? (event.order === 1 ? -1 : 1) : sortDesc;
      }
    });

    return [...items];
  }
}
