import { OnInit, OnDestroy, ViewChild, ChangeDetectorRef, Directive } from "@angular/core";
import { FormBuilder, FormArray } from "@angular/forms";
import { ActivatedRoute } from "@angular/router";
import { Location } from "@angular/common";
import { EntityCollectionServiceBase, DefaultDataService, EntityOp } from "@ngrx/data";
import { Actions } from "@ngrx/effects";
import { takeUntil } from "rxjs/operators";
import { FileUploader, FileUploaderOptions, FileItem } from "ng2-file-upload";

import { BaseFormComponent } from "./base-form.component";
import { MessageService, Error, BitMessage, BitNGUploadComponent, FormType, CustomValidators } from "../../../shared";

/**
 * Componente base que implementa un formulario para insertar, modificar, eliminar y/o consultar un elemento de tipo T
 * y que además tendrá la posibilidad de subida de ficheros
 */
@Directive()
export abstract class BaseFormUploadComponent<T> extends BaseFormComponent<T> implements OnInit, OnDestroy {
  uploader: FileUploader;
  //parametros para personalizar el uploader
  hasDropZone: boolean = true;
  hasQueue: boolean = true;
  hasSingleUpload: boolean = true;
  hasMultipleUpload: boolean = true;
  hasProgressBar: boolean = false; //inicialmente no hay barra de progreso. En el método setCanUpload podemos cambiar eso
  deleteItemAfterUpload: boolean = true;
  queueLimit: number = 10;
  itemSizeLimit: number = 100000000;
  sizeDropZone: number = 4;
  sizeQueue: number = 8;
  canUpload: boolean; //este parametro dependerá de si estamos creando un nuevo usuario (false) o si estamos editándolo (true)

  uploadedFiles: FileItem[]; //ficheros que se han subido o asociados al objeto
  //referencia al uploader. La necesitamos para poder reiniciar la cola cuanto el usuario opte por subir los ficheros con el método save de este método
  @ViewChild("upload")
  uploadComponent: BitNGUploadComponent;
  urlUpload: string;

  constructor(
    protected activatedRoute: ActivatedRoute,
    protected messageService: MessageService,
    protected formBuilder: FormBuilder,
    protected location: Location,
    protected actions$: Actions,
    protected entityService: EntityCollectionServiceBase<T>,
    protected dataService: DefaultDataService<T>,
    protected changeDetector: ChangeDetectorRef
  ) {
    super(activatedRoute, messageService, formBuilder, location, actions$, entityService);
    this.uploadedFiles = [];
  }

  ngOnInit() {
    super.ngOnInit();
    this.createUploader();
    this.setCanUpload();
  }

  /** crea un uploader */
  createUploader() {
    this.uploader = new FileUploader(this.getFileUploaderOptions());
  }

  /** options que definirán nuestro uploader */
  getFileUploaderOptions(): FileUploaderOptions {
    return {
      url: this.urlUpload,
      queueLimit: this.queueLimit, //TODO: poner propiedad
      headers: [{ name: "Accept", value: "application/json" }],
      maxFileSize: this.itemSizeLimit, ////TODO: poner propiedad permite un tamaño máximo de fichero en bytes
      //authTokenHeader: "Authorization",  //si el método para subida al servidor está protegido podemos setear el token según corresponda
      //authToken: "Bearer " + token"
      disableMultipart: false // 'DisableMultipart' must be 'true' for formatDataFunction to be called.
      // formatDataFunctionIsAsync: true,
      // formatDataFunction: async item => {   //si no queremos utilizar la forma multipart si no enviar en el JSON todo el mega churro de bytes
      //   return new Promise((resolve, reject) => {
      //     resolve({
      //       name: item._file.name,
      //       length: item._file.size,
      //       contentType: item._file.type,
      //       date: new Date()
      //     });
      //   });
      // }
    };
  }

  /** Setea si el componente upload puede hacer uplodas individuales. Si no nos gusta este comportamiento podemos extender este método */
  setCanUpload() {
    if (this.isEditMode) {
      this.hasProgressBar = true;
      this.canUpload = true; //solo si estamos editando un objeto permitiemos subidas individuales
    }
  }

  /** Evento que se dispara dsde el uploader cuando un fichero individual ha sido subido al servidor de manera exitosa */
  onSuccess(event) {
    console.log(event);
    this.uploadedFiles.push(event.item);
    this.messageService.sendNotification(new BitMessage("Subida fichero", "Fichero subido satisfactoriamente"));
  }

  /** Evento que se dispara dsde el uploader cuando un fichero individual NO ha sido subido al servidor de manera exitosa */
  onError(event) {
    let error: string = `error uploaded ${event.status}`;
    this.messageService.sendError(new Error(event.item.url, event.response, event.response, error));
  }

  //sobreescribimos el método save estandar porque necesitamos setear los ficheros que puedan subirse
  save() {
    let fileExists = this.uploader.queue.length > 0;
    //primero averiguamos si hay ficheros para subir. En ese caso, mapeamos en el form
    if (!fileExists) {
      super.save(); //si no los tiene es un save normalito
    } else {
      //si hay ficheros a anexar vamos por otro lado
      //primero miramos los errores del formulario
      if (!this.myForm.valid) {
        let errores = CustomValidators.validateForm(this.myForm);
        this.messageService.sendError(
          new Error(null, "Error al formulari", "Error en el formulari", null, null, true, "", errores)
        );
      } else {
        //nos guardamos los ficheros a subir en filesToUpload
        let filesToUpload = new Array();
        for (let j = 0; j < this.uploader.queue.length; j++) {
          let fileItem = this.uploader.queue[j]._file;
          filesToUpload.push(fileItem);
        }
        //seteamos el object que irá al server. De momento sin los ficheros
        const newObject: T = {
          ...this.object,
          ...this.myForm.value
        };
        if (this.preSave()) {
          let service; //contendrá la función que nos hará el insert o el update. Es responsabilidad del componente que exitiende a este
          //el proporcionarla implementando ambos métodos
          if (this.formType == FormType.Insert) {
            service = this.insertAndUploadService(newObject, filesToUpload);
          } else {
            service = this.updateAndUploadService(newObject, filesToUpload);
          }
          //cuando se ejecute la función de insert o update
          service.pipe(takeUntil(this.destroy$)).subscribe(result => {
            this.entityService.upsertOneInCache(result); //actualizamos la caché del entityData en el store
            this.entityService.createAndDispatch(EntityOp.COMMIT_ALL); //comitamos en el store los cambios
            this.postSave(result); //cosas a realizar tras guardar
          });
        }
      }
    }
  }

  /** sobreescribimos el método postSave porque necesitamos resetear el campo de upload del template y además debemos resetear el uploader
   * con la opción para permitir subir ficheros de manera indiviual
   */
  postSave(result: T) {
    super.postSave(result);
    this.setCanUpload();
    this.uploadComponent.removeQueue();
    this.uploader.setOptions(this.getFileUploaderOptions()); //reseteamos las opciones del FileUploader. Suele cambiar la URL de subida
    this.changeDetector.detectChanges(); //angular a veces es caprichoso con la detección de cambios y necesitamos avisarle que la haga
  }

  /** Debe retornar la función que permite insertar el objeto y subir los documentos en el servidor */
  abstract insertAndUploadService(obj: T, filesToUpload: FileItem[]);

  /** Debe retornar la función que permite modificar el objeto y subir los documentos en el servidor */
  abstract updateAndUploadService(obj: T, filesToUpload: FileItem[]);
}
