import { HostListener, OnInit, OnDestroy, Directive } from "@angular/core";
import { FormBuilder, FormGroup } from "@angular/forms";
import { ActivatedRoute } from "@angular/router";
import { Location } from "@angular/common";
import { createEffect, Actions } from "@ngrx/effects";
import { EntityCollectionServiceBase } from "@ngrx/data";
import { Observable, from, of, merge, Subject } from "rxjs";
import { map, filter, takeUntil } from "rxjs/operators";

import { getInitToolbar, getToolbarObjectLoadedDefault } from "./../../../base/models";
import {
  PageStatus,
  FormType,
  MessageService,
  CustomValidators,
  BitToolbarComponent,
  BitMessage,
  Error
} from "../../../shared";
import { BaseComponent } from "./base.component";

/**
 * Componente que implementa un formulario para insertar, modificar, eliminar y/o consultar un elemento de tipo T usando el framework
 * ngrx-data
 */
@Directive()
export abstract class BaseFormComponent<T> extends BaseComponent implements OnInit, OnDestroy {
  myForm: FormGroup;
  object: T;
  status$: Observable<PageStatus>;
  toolbar$: Observable<{}>;
  breadcrumb: any[] = []; // contiene la miga de pan de nuestro componente
  formType: FormType;
  soloLectura: boolean; //sera true dependiendo de las circunstacias

  /*tabs properties*/
  myTabs: any = {}; // pestañas definidas en el componente (ver registerTab() e initTabs())
  initialTab: string; // pestaña cargada inicialmente (indicada por parámetro)
  loadedTab: any = {}; // pestañas con datos cargados (todas se cargan de forma perezosa)
  currentTab: number;

  destroy$: Subject<boolean> = new Subject<boolean>();

  constructor(
    protected activatedRoute: ActivatedRoute,
    protected messageService: MessageService,
    protected formBuilder: FormBuilder,
    protected location: Location,
    protected actions$: Actions,
    protected entityService: EntityCollectionServiceBase<T>
  ) {
    super();
    console.log("base-form creado");
    this.toolbar$ = this.createToolbar();
    this.breadcrumb = this.createBreadcrumb(FormType.Insert);
    this.myForm = this.createFormGroup();
  }

  ngOnInit(): void {
    let id: number = +this.activatedRoute.snapshot.paramMap.get("id");
    this.object = this.createObject(id);
    if (this.object) {
      this.setModeUpdate(this.object);
      this.updateFormValues();
    } else {
      this.setModeInsert();
    }
    this.updatePageStatus();
  }

  ngOnDestroy() {
    this.destroy$.next(true);
    this.destroy$.unsubscribe();
  }

  /** Primer valor de la toolbar al crear el objeto BaseForm */
  createToolbar(): Observable<{}> {
    return of(getInitToolbar());
  }

  /** Primer valor del objeto que se mapeará en el formulario asociado.  */
  abstract createObject(id: number): T;

  /** Construcción del FormGroup que tendrá asociado el formulario */
  abstract createFormGroup(): FormGroup;

  /** Valor del objeto breadcrumb dependiendo de si el componente va a insertar un nuevo objeto o actualizarlo */
  abstract createBreadcrumb(formType: FormType): any[];

  /** Establece el componente para editar un objeto ya existente */
  setModeUpdate(object) {
    this.object = object;
    this.formType = FormType.Update;
    this.breadcrumb = this.createBreadcrumb(FormType.Update);
    this.toolbar$ = of(getToolbarObjectLoadedDefault());
  }

  /** Establece el componente para insertar un nuevo objeto */
  setModeInsert() {
    this.formType = FormType.Insert;
  }

  /** Actualiza los valores del FormGroup con el valor actual almacenado en object */
  updateFormValues() {
    this.myForm.patchValue({ ...this.object });
  }

  /** Actualiza el page status en función de tres observables. El estado de envío al servidor, el posible error producido y el éxito de la operación*/
  updatePageStatus() {
    this.status$ = merge(
      this.entityService.loading$.pipe(
        map(loading => {
          if (loading) {
            return PageStatus.Searching;
          }
        })
      ),
      createEffect(
        () =>
          this.actions$.pipe(
            map(action => action.type),
            filter(actionType => actionType.indexOf("error") != -1),
            map(action => PageStatus.Error)
          ),
        { dispatch: false }
      ),
      createEffect(
        () =>
          this.actions$.pipe(
            map(action => action.type),
            filter(actionType => actionType.indexOf("success") != -1),
            map(action => PageStatus.FinishSearch)
          ),
        { dispatch: false }
      )
    );
  }

  /** Limpia el formulario. */
  abstract cleanForm(): void;

  /** Devuelve true si el formulario edita un elemento existente, o false si inserta un nuevo elemento. */
  get isEditMode(): boolean {
    return this.formType === FormType.Update;
  }
  /**
   * Devuelve true si hay algún valor cambiado en la ficha.
   * Puede ser por un campo del formulario (this.myForm.dirty) o por modificaciones en elementos relacionados.
   */
  get isDataChanged(): boolean {
    return this.myForm && this.myForm.dirty;
  }

  /** Devuelve true si el formulario es válido (campos validados automáticamente, elementos relacionados, etc.). */
  get isFormValid(): boolean {
    return this.myForm && this.myForm.valid;
  }

  /** Evalúa si el usuario puede desactivar/cerrar el componente sin perder cambios todavía no guardados. */
  canDeactivate(): Observable<boolean> | boolean {
    console.log("canDeactivate called");
    if (this.isDataChanged) {
      //si hay cambios, mostramos la ventana de diálogo
      let p = this.messageService.confirm(
        "Hi ha canvis pendents de desar. Està segur de que vol sortir de la plana i perdre-los?"
      );
      let o = from(p);
      return o;
    } else {
      // si no hay cambios, no la mostramos
      console.log("canDeactivate true");
      return true;
    }
  }

  /** Controla si el usuario sale de la página abrúptamente */
  @HostListener("window:beforeunload")
  unloadWithUnsavedChanges($event: any) {
    console.log("unloadWithUnsavedChanges called");
    let msg = "Hi ha canvis pendents de desar. Està segur de que vol sortir de la plana i perdre-los?";
    if (this.isDataChanged) {
      ($event || window.event).returnValue = msg;
      return msg;
    }
    return undefined;
  }

  /** Método que se dispara cuando algún botón de la toolbar es pulsado */
  onToolbarButtonPressed(button: string) {
    console.log("Button pressed:" + button);
    if (button === BitToolbarComponent.SAVE_BUTTON) {
      this.save();
    } else if (button === BitToolbarComponent.DELETE_BUTTON) {
      this.delete();
    } else if (button === BitToolbarComponent.PRINT_BUTTON) {
      this.handlePrint();
    } else if (button === BitToolbarComponent.HELP_BUTTON) {
      this.help();
    } else {
      this.otherAction(button);
    }
  }

  /** Muestra la ayuda asociada a la pantalla. */
  help(): void {}

  /** el componente toolbar nos comunica que se ha pulsado un botón */
  handlePrint() {
    window.print();
  }

  /** contiene toda la lógica para enviar el formulario al servidor*/

  /** función útil que los componentes pueden extender para hacar algo antes de enviar el formulario al servidor */
  preSave(): boolean {
    return true;
  }

  save(): void {
    console.log("save called");
    if (!this.myForm.valid) {
      let errores = CustomValidators.validateForm(this.myForm);
      this.messageService.sendError(
        new Error(null, "Error al formulari", "Error en el formulari", null, null, true, "", errores)
      );
    } else {
      const newObject = {
        ...this.object,
        ...this.myForm.value
      };
      if (this.preSave()) {
        console.log("presave");
        if (this.isEditMode) {
          console.log("update object");
          this.entityService
            .update(newObject)
            .pipe(takeUntil(this.destroy$))
            .subscribe(object => this.postSave(object));
        } else {
          console.log("insert object");
          this.entityService
            .add(newObject)
            .pipe(takeUntil(this.destroy$))
            .subscribe(object => this.postSave(object));
        }
      }
    }
  }

  /** función para qué hacer después de insertar o modificar */
  postSave(object: any) {
    this.setModeUpdate(object); //actualizamos nuestra copia local con lo que viene del servidor
    this.myForm.patchValue({
      ...this.object
    });
    this.myForm.markAsPristine(); // el formulario vuelve a estar sincronizado con el servidor
    this.myForm.markAsUntouched();
    this.messageService.sendNotification(new BitMessage("Informació", "Les dades s'han desat correctament."));
  }

  preDelete() {}

  /** contiene la lógica para eliminar el objeto representado en el formulario */
  delete(): void {
    this.preDelete();
    let p = this.messageService.confirm(
      "Està segur que vol eliminar aquest registre? Aquesta acciò no es podrà desfer?"
    );
    let local = this;
    p.then(function(confirmacion) {
      if (confirmacion) {
        local.entityService
          .delete(local.object)
          .pipe(takeUntil(local.destroy$))
          .subscribe(() => local.postDelete());
      }
    });
  }

  postDelete() {
    this.messageService.sendNotification(new BitMessage("Informació", "Les dades s'han eliminat correctament."));
    this.goBack();
  }

  /** Cualquier otra accion que no sea la típica de guardar o eliminar */
  otherAction(action: string): void {}

  /** Cierra el formulario y vuelve a la pantalla anterior. */
  goBack(): void {
    this.location.back();
  }

  isPreserveQueryParams() {
    return false;
  }

  /** Información de debug. */
  get diagnostic() {
    return JSON.stringify(this.object);
  }

  // Métodos relacionados con pestañas (tabs) //

  /** función para hacer algo si queremos al cambiar de pestaña */
  changeTab(e) {}

  /** Registra una pestaña de visualización de información. */
  registerTab(index: number, id: string) {
    this.myTabs[id] = index;
  }

  // inicialización de pestañas (si hay)
  initTabs() {
    this.currentTab = 0;
    this.initialTab = this.activatedRoute.snapshot.params["tab"];
    for (let tab in this.myTabs) {
      let isInitial = this.isInitialTab(tab);
      this.loadedTab[tab] = isInitial;
      if (isInitial) {
        this.currentTab = this.myTabs[tab];
      }
    }
  }

  /** Indica si la pestaña correspondiente es la inicial. */
  isInitialTab(tab: string) {
    return this.initialTab === tab;
  }

  /** Indica si la pestaña correspondiente está cargada. */
  isLoadedTab(tab: string) {
    return this.loadedTab[tab];
  }
}
