import { ActivatedRoute, Router } from "@angular/router";
import { Actions, createEffect } from "@ngrx/effects";
import { EntityCollectionServiceBase } from "@ngrx/data";
import { Observable, of, merge } from "rxjs";
import { map, filter, takeUntil } from "rxjs/operators";
import * as querystring from "querystring";

import { PageStatus, BitToolbarComponent } from "../../../shared";
import { TableViewService } from "../../../shared/service/tableview.service";
import { BaseFilter } from "../../shared/base.filter";
import { makeToolbarList, FactoryCreateNewFilterFromFilter } from "../../../base/models";
import { BaseComponent } from "./base.component";
import { Directive } from "@angular/core";
import { objectoToQueryParameters } from "../../shared/utils";

/**
 * Componente base para listar objectos de tipo T utilizando filtros de búsqueda de tipo F mediante el framework ngrx-data
 */
@Directive()
export abstract class BaseListComponent<T, F extends BaseFilter> extends BaseComponent {
  breadcrumb: any[];
  items$: Observable<T[]>;
  totalItems$: Observable<number>;
  status$: Observable<PageStatus>;
  toolbar$: Observable<{}>;
  filter: F;
  customToolbarActions: any = {};

  constructor(
    protected actions$: Actions,
    protected tableViewService: TableViewService,
    protected activatedRoute: ActivatedRoute,
    protected router: Router,
    protected entityService: EntityCollectionServiceBase<T>,
    protected newFilterFunc: FactoryCreateNewFilterFromFilter<F>
  ) {
    super();
    this.breadcrumb = this.createBreadcrumb();
    this.toolbar$ = this.createToolbar(false, false);
  }

  ngOnInit() {
    console.log("ngOnInit called");
    this.updateToolbar();
    this.updateItems();
    this.updatePageStatus();
    this.updateFilter();
    this.initSearch();
  }

  ngOnDestroy() {
    this.destroy$.next(true);
    this.destroy$.unsubscribe();
  }

  /** Primer valor de la toolbar al crear el objeto BaseList */
  createToolbar(readOnly: boolean, enable: boolean): Observable<{}> {
    return of(makeToolbarList(readOnly, enable));
  }

  /** Decide lanzar la búsqueda al crear el componente o no en función de si en la URL de carga del componente se ha especificado el parámetro
   * query a true
   */
  initSearch() {
    let query = this.activatedRoute.snapshot.params["query"];
    console.log(`parametro query con valor ${query}`);
    if (query == "true") {
      this.search(this.filter);
    }
  }

  abstract createBreadcrumb(): any[];

  abstract gotoForm(data?: any);

  /**actualización del toolbar en función de si hay elementos cargados tras la búsqueda o no hay elementos*/
  updateToolbar() {
    this.entityService.entities$.pipe(takeUntil(this.destroy$)).subscribe((items) => {
      this.toolbar$ = this.createToolbar(false, items && items.length > 0);
    });
  }

  /**actualización de los items y la cuenta de items*/
  updateItems() {
    this.items$ = this.entityService.entities$;
    this.totalItems$ = this.entityService.count$;
  }

  /**actualiza el page status en función de tres observables. El estado de envío al servidor, el posible error producido y el éxito de la operación*/
  updatePageStatus() {
    this.status$ = merge(
      this.entityService.loading$.pipe(
        map((loading) => {
          if (loading) {
            return PageStatus.Searching;
          }
        })
      ),
      createEffect(
        () =>
          this.actions$.pipe(
            map((action) => action.type),
            filter((actionType) => actionType.indexOf("error") != -1),
            map((action) => PageStatus.Error)
          ),
        { dispatch: false }
      ),
      createEffect(
        () =>
          this.actions$.pipe(
            map((action) => action.type),
            filter((actionType) => actionType.indexOf("success") != -1),
            map((action) => PageStatus.FinishSearch)
          ),
        { dispatch: false }
      )
    );
  }

  /**actualización del filtro de búsqueda*/
  updateFilter() {
    this.entityService.filter$.pipe(takeUntil(this.destroy$)).subscribe((filter) => {
      if (filter) {
        this.filter = this.newFilterFunc(JSON.parse(filter));
        console.log("filter recibido", this.filter);
      }
    });
  }

  /** el componente toolbar nos comunica que se ha pulsado un botón */
  onToolbarButtonPressed(button: string) {
    console.log("Button pressed:" + button);
    if (this.customToolbarActions[button]) {
      this.customToolbarActions[button]();
    } else if (button === BitToolbarComponent.PRINT_BUTTON) {
      this.handlePrint();
    } else if (button === BitToolbarComponent.EXPORT_BUTTON) {
      this.handleExport();
    } else if (button === BitToolbarComponent.NEW_BUTTON) {
      this.handleNew();
    } else {
      this.otherAction(button);
    }
  }

  /** el componente toolbar nos comunica que se ha pulsado un botón */
  handlePrint() {
    window.print();
  }

  /** Cuando tenemos que exportar el listado */
  handleExport() {
    if (this.tableViewService) {
      this.tableViewService.sendExportMessage(); //enviamos un mensaje a un observable para que el tableview active el export
    } else {
      alert("Exportació no implementada per aquest llistat");
    }
  }

  /** Cuando tenemos que nagevar al formulario para crear un nuevo elemento*/
  handleNew() {
    this.gotoForm();
  }

  /**cualquier otra accion que no sea la típica de guardar o eliminar */
  otherAction(button: string): void {
    console.warn(`otherAction ${button} pressed`);
  }

  /** Se invocará cuando el componente paginator hijo nos comunique que el usuario ha cambiado el número de elementos a mostrar por página */
  onNewPageSize(pageSize: number): void {
    console.log(`onNewPageSize ${pageSize}`);
    this.filter.pagina = 1; // reseteamos a la página inicial para evitar comportamienos extraños
    this.filter.resultados = pageSize;
    this.entityService.setFilter(JSON.stringify(this.filter));
    this.search(this.filter);
  }

  /** Se invocará cuando el componente paginator hijo nos comunique que el usuario ha cambiado el número de página a mostrar */
  onPageChange(page: number): void {
    console.log(`onPageChange ${page}`);
    this.filter.pagina = page;
    this.entityService.setFilter(JSON.stringify(this.filter));
    this.search(this.filter);
  }

  /** Lanza la búsqueda contra el servidor */
  search(filter: F) {
    console.log("search called ", filter);
    this.entityService.setFilter(JSON.stringify(filter));
    //indicamos al store que la lista no se ha cargado porque la queremos volver a cargar. Esto evita efectos raros
    this.entityService.setLoaded(false);
    let queryParams: string = objectoToQueryParameters(filter.allValues());
    this.entityService.getWithQuery(queryParams);
  }

  /** Resetea el filtro de búsqueda */
  reset(filter: F) {
    console.log("reset called ", filter);
    this.entityService.setFilter(JSON.stringify(this.filter));
  }
}
