import { Input, EventEmitter, Output, SimpleChange, OnInit, OnDestroy, Directive } from "@angular/core";
import { EntityCollectionServiceBase, EntityOp, makeSuccessOp } from "@ngrx/data";
import { Observable, of } from "rxjs";

import { MessageService, FormType, Item, BitToolbarComponent } from "../../../shared";
import { BaseComponent } from "../../../base/component/ngrx-data/base.component";
import { makeToolbarSubList } from "../../../base/models/base-sublist.models";
import { BaseDefaultDataService } from "../../shared/base-default-data-service";
import { COMMIT } from "@ngrx/store-devtools/src/actions";

/**
 * Componente basesublist de listado de entidades.
 */
@Directive()
export abstract class BaseSubListComponent<T> extends BaseComponent implements OnInit, OnDestroy {
  displayInnerForm: boolean; // si tenemos que mostrar el form para editar el objeto en linea o no
  modeInnerForm: FormType; // modo en que mostramos el inner form
  itemSelected$: Observable<T>; // objeto a enviar al InnerForm
  selectedItems: number[]; // elementos checkeados
  items$: Observable<T[]>;
  toolbar$: Observable<{}>;

  @Input()
  parent: Item; // objeto padre que es dueño de la lista
  @Input()
  parentField: string; // propiedad de los elementos que establece la relación con su padre
  @Input()
  initLoad: boolean; // cuando sea true la lista se cargará
  @Input()
  itemsPreloaded: Item[]; // lista precargada de elementos que el padre nos inyecta
  @Input()
  femenino: boolean = false; // si tenemos que mostrar los textos en femenino o no (p. ej., Nou vs Nova)
  @Input()
  forceReload: EventEmitter<any>; // opcional, eventos a esta cola nos dicen que debemos forzar una recarga
  @Input()
  reloadFromServer: boolean = true; //indica si una vez añadido un elemento, llamaos al servidor para recargar la lista. Por defecto sí lo hace.
  @Output()
  onChange = new EventEmitter<boolean>();

  constructor(
    protected messageService: MessageService,
    protected entityService: EntityCollectionServiceBase<T>,
    protected dataService: BaseDefaultDataService<T>
  ) {
    super();
  }

  ngOnInit() {
    this.toolbar$ = this.createToolbar(false);
    this.items$ = this.getEntities();
    this.itemSelected$ = of({} as T);
    this.modeInnerForm = FormType.Insert;
    this.displayInnerForm = false;
    this.loadItems();
  }

  ngOnDestroy() {
    super.ngOnDestroy();
  }

  /** Controlamos si la propiedad initLoad cambia a true para cargar la lista de documentos */
  ngOnChanges(changes: { [propKey: string]: SimpleChange }) {
    for (let propName in changes) {
      if (propName === "forceReload") {
        this.forceReload &&
          this.forceReload.subscribe(
            function(x) {
              this.loadItems();
            }.bind(this)
          );
      }
    }
  }

  createToolbar(enableDelete: boolean): Observable<{}> {
    return of(makeToolbarSubList(enableDelete));
  }

  /** Recupera el listado de elementos a mostrar en esta sublista */
  abstract getEntities(): Observable<T[]>;

  /** Abre el formulario para editar el elemento seleccionado */
  abstract gotoForm();

  /** Carga los elementos a mostrar en el listado desde el servidor */
  abstract loadItems();

  //Nos presionan algún botón del toolbar
  onToolbarButtonPressed(button: string) {
    console.log(`Button pressed: ${button}`);
    if (button === BitToolbarComponent.NEW_BUTTON) {
      this.gotoForm();
    } else if (button === BitToolbarComponent.DELETE_BUTTON) {
      this.confirmDeleteItems();
    } else {
      this.otherAction(button);
    }
  }

  /** Cualquier otra accion que no sea la típica de guardar o eliminar */
  otherAction(action: string): void {
    console.warn(`otherAction ${action} pressed`);
  }

  /** Elimina un elemento de la lista cuando venimos del formulario emergente */
  deleteItem(id: number): void {
    console.log(`deleteItem called for id ${id}`);
    this.entityService.delete(id).subscribe(() => (this.displayInnerForm = false));
  }

  /** Inserta o actualiza el documento */
  saveItem(item: T): void {
    console.log(`saveItem called for item ${item}`);
    if (this.modeInnerForm == FormType.Insert) {
      this.entityService.add(item).subscribe(() => (this.displayInnerForm = false));
    } else {
      this.entityService.update(item).subscribe(() => (this.displayInnerForm = false));
    }
  }

  /** nos piden cerrar el formulario */
  closeForm(event) {
    this.displayInnerForm = false;
  }

  /** Solicita confirmación para eliminar los elementos seleccionados */
  confirmDeleteItems() {
    let p = this.messageService.confirm(
      "Està segur que vol eliminar els elements seleccionats? Aquesta acciò no es podrà desfer."
    );
    let local = this;
    p.then(function(confirmacion) {
      if (confirmacion) {
        local.deleteItems();
      }
    });
  }

  /** Elimina los documentos seleccionados en el listado al pulsar el botón de elimina de la toolbar */
  deleteItems(): void {
    console.log("delete items", this.selectedItems);
    //primero borramos de la cache siendo optimistas. Se creará un stateChange pendiente de comitar o no
    this.entityService.removeManyFromCache(this.selectedItems);
    this.dataService.deleteEntities(this.selectedItems).subscribe(
      () => this.postDelete(true),
      error => this.postDelete(false)
    );
  }

  /** Función para hacer algo después de haber eliminado los elementos seleccionados */
  postDelete(result: boolean) {
    console.log("postDelete llamado", result);
    if (result) {
      this.entityService.createAndDispatch(EntityOp.COMMIT_ALL); //comitamos en el store los cambios
      this.toolbar$ = this.createToolbar(false);
      this.selectedItems = [];
    } else {
      this.entityService.createAndDispatch(EntityOp.UNDO_ALL); //deshacemos los cambios porque tenemos un error
    }
  }

  /** Nos checkean algún item del listado */
  itemsChecked(selectedItems: number[]) {
    console.log("items selected", selectedItems);
    this.selectedItems = selectedItems;
    if (selectedItems !== null && selectedItems.length > 0) {
      this.toolbar$ = this.createToolbar(true);
    } else {
      this.toolbar$ = this.createToolbar(false);
    }
  }
}
