import {
  Component,
  ChangeDetectionStrategy,
  OnInit,
  Input,
  EventEmitter,
  Output,
  ViewChild,
  ChangeDetectorRef
} from "@angular/core";

import { FileUploader, FileItem, ParsedResponseHeaders } from "ng2-file-upload";
import { GeneralUtils } from "../service/general-utils.service";

@Component({
  selector: "bit-ng-upload",
  templateUrl: "bit-ng-upload.component.html",
  styleUrls: ["bit-ng-upload.component.css"],
  changeDetection: ChangeDetectionStrategy.OnPush
})
export class BitNGUploadComponent implements OnInit {
  @Input()
  uploader: FileUploader;
  @Input()
  hasDropZone: boolean;
  @Input()
  hasQueue: boolean;
  @Input()
  hasSingleUpload: boolean;
  @Input()
  hasMultipleUpload: boolean;
  @Input()
  deleteItemAfterUpload: boolean;
  @Input()
  canUpload: boolean;
  @Input()
  hasProgressBar: boolean;
  @Input()
  sizeDropZone: number;
  @Input()
  sizeQueue: number;

  @Output()
  onSuccess: EventEmitter<any> = new EventEmitter<any>();
  @Output()
  onError: EventEmitter<any> = new EventEmitter<any>();

  @ViewChild("fileRef")
  fileInput: any;

  hasBaseDropZoneOver: boolean;

  constructor(private generalUtils: GeneralUtils, private changeDetector: ChangeDetectorRef) {}

  ngOnInit(): void {
    this.uploader.onSuccessItem = (
      item: FileItem,
      response: string,
      status: number,
      headers: ParsedResponseHeaders
    ) => {
      console.log("success upload file");
      this.stopLoading();
      this.onSuccess.emit({ item, response, status, headers });
      if (this.deleteItemAfterUpload) {
        item.remove();
      }
    };
    this.uploader.onErrorItem = (item: FileItem, response: string, status: number, headers: ParsedResponseHeaders) => {
      console.log("error upload file");
      this.stopLoading();
      this.onError.emit({ item, response, status, headers });
    };
    this.uploader.onProgressItem = (item: FileItem, progress: any) => {
      console.log(progress);
      this.changeDetector.detectChanges();
    };

    this.hasBaseDropZoneOver = false;
  }

  /** Arrancará el spinner en el caso de que la barra de progreso esté desactivada*/
  startLoading() {
    if (!this.hasProgressBar) {
      this.generalUtils.startLoading();
    }
  }

  /** Parará el spinner en el caso de que la barra de progreso esté desactivada*/
  stopLoading() {
    if (!this.hasProgressBar) {
      this.generalUtils.stopLoading();
    }
  }

  /** Sube todos los ficheros en cola al servidor */
  uploadAll() {
    this.startLoading();
    this.uploader.uploadAll();
  }

  /** Elimina la cola de ficheros y deselecciona el fichero que pudiera estar seleccionado en el single input file */
  removeQueue() {
    this.stopLoading();
    this.uploader.clearQueue();
    this.changeDetector.detectChanges();
    this.removeInputFile();
  }

  /** Cancela la subida de ficheros al servidor */
  cancelQueue() {
    this.stopLoading();
    this.uploader.cancelAll();
  }

  /** Sube un fichero al servidor de manera individual */
  uploadFile(item) {
    this.startLoading();
    item.upload();
  }

  /** Elimina de la cola un fichero en concreto y en el caso de que coincida con el seleccionado en el single input file, tambien
   * lo deselecciona
   */
  removeFile(item) {
    this.stopLoading();
    if (this.fileInput && this.fileInput.nativeElement.value.indexOf(item.file.name) != -1) {
      this.removeInputFile();
    }
    item.remove();
  }

  /** Cancela la subida de un fichero de manera individual */
  cancelFile(item) {
    this.stopLoading();
    item.cancel();
  }

  /** Deselecciona el fichero seleccionado en el input file */
  removeInputFile() {
    if (this.fileInput) this.fileInput.nativeElement.value = "";
  }

  /** Evento que se dispara cuando arrastran un fichero a la zona de drag&drop */
  public fileOverBase(e: any): void {
    this.hasBaseDropZoneOver = e;
  }
}
