import { ActivatedRouteSnapshot, RouterStateSnapshot, Params } from "@angular/router";
import { ActionReducerMap, createFeatureSelector, ActionReducer, MetaReducer, createSelector } from "@ngrx/store";
import * as fromRouter from "@ngrx/router-store";

export const ROUTER_STATE: string = "routerState";

export function saveRouterStateIntoLocalStorage(routerState) {
  localStorage.setItem(ROUTER_STATE, JSON.stringify(routerState));
}

export function recoverRouterStateFromLocalStorage(): RouterStateUrl {
  let routerStateStorage = localStorage.getItem(ROUTER_STATE);
  let routerStateUrl: RouterStateUrl;
  if (routerStateStorage != null) {
    routerStateUrl = JSON.parse(routerStateStorage);
  }
  return routerStateUrl;
}

// representación de la URL en la que nos encontramos como fuente de la verdad
export interface RouterStateUrl {
  url: string;
  queryParams: Params;
  params: Params;
}

// necesitamos incluir routerReducer como state. El paquete @ngrx/router-store nos da la representación del estado
// mediante RouterReducerState
export interface AppState {
  routerReducer: fromRouter.RouterReducerState<RouterStateUrl>;
}

export const reducers: ActionReducerMap<AppState> = {
  routerReducer: fromRouter.routerReducer
};

// selector al estate principal
export const getRouterReducerState = createFeatureSelector<fromRouter.RouterReducerState<RouterStateUrl>>(
  "routerReducer"
);

// selector para recuperar el estado que nos dará la url, los queryParams y los params
export const getRouterStateUrl = createSelector(getRouterReducerState, (state) => state.state);

/*
 custom serializer que nos permite obtener la representación del estado de la URL donde nos encontramos en cada momento
 Registrando estsa clase conseguiermos que el RouterStore quedará escuchando eventos del Router, por lo que cada vez que naveguemos 
 y accedamos a otra URL, esta clase se ejecutará para obtener esta nueva representación
*/
export class CustomSerializer implements fromRouter.RouterStateSerializer<RouterStateUrl> {
  //Este método nos transforma el estado del router de Angular a nuestra propia estructura
  serialize(routerSate: RouterStateSnapshot): RouterStateUrl {
    const { url } = routerSate; //same as url = routerSate.url;
    const { queryParams } = routerSate.root;

    let state: ActivatedRouteSnapshot = routerSate.root;
    //El router es un árbol de estados sobre el que podemos iterar
    while (state.firstChild) {
      state = state.firstChild;
    }
    const { params } = state;
    //este objeto se linkará a nuestro árbol de estados
    return { url, queryParams, params };
  }
}

/** An example of metareducer function that logs when an action is dispatched */
export function logger(reducer: ActionReducer<any>): ActionReducer<any> {
  return (state, action) => {
    // console.log("state before: ", state);
    // console.log("action", action);

    //We're are continue doing the chain, returning the next reducer function that must be called
    return reducer(state, action);
  };
}

export const metaReducers: MetaReducer<AppState>[] = [];
