import { Injectable } from "@angular/core";
import { Router } from "@angular/router";
import { Actions, ofType, createEffect } from "@ngrx/effects";
import { switchMap, map, catchError, tap } from "rxjs/operators";
import { of } from "rxjs";

import { Usuario } from "../../../../common/data/usuario";
import * as usuarioFormActions from "../actions/usuario-form.action";
import { MessageService, FormType } from "../../../../shared";
import { BitMessage } from "../../../../shared/data/bitmessage";
import { UsuarioService } from "../../shared/usuario.service";

@Injectable()
export class UsuarioFormEffects {
  constructor(
    private actions$: Actions,
    private usuarioService: UsuarioService,
    private messageService: MessageService,
    private router: Router
  ) {}

  loadUsuario$ = createEffect(() =>
    this.actions$.pipe(
      ofType(usuarioFormActions.loadUsuario),
      map(action => action.id),
      switchMap(id => {
        return this.usuarioService.get<Usuario>(id).pipe(
          map(usuario => usuarioFormActions.loadUsuarioSuccess({ usuario })),
          catchError(error => of(usuarioFormActions.loadUsuarioFail({ error })))
        );
      })
    )
  );

  insertUsuario$ = createEffect(() =>
    this.actions$.pipe(
      ofType(usuarioFormActions.insertUsuario),
      map(action => action.usuario),
      switchMap(usuario => {
        return this.usuarioService.save<Usuario>(usuario, FormType.Insert, null).pipe(
          tap(() => {
            this.messageService.sendNotification(new BitMessage("Informació", "Les dades s'han desat correctament."));
          }),
          map(usuario => usuarioFormActions.insertUsuarioSuccess({ usuario })),
          catchError(error => of(usuarioFormActions.insertUsuarioFail({ error })))
        );
      })
    )
  );

  updateUsuario$ = createEffect(() =>
    this.actions$.pipe(
      ofType(usuarioFormActions.updateUsuario),
      map(action => action.usuario),
      switchMap(usuario => {
        return this.usuarioService.save<Usuario>(usuario, FormType.Update, null).pipe(
          tap(() => {
            this.messageService.sendNotification(new BitMessage("Informació", "Les dades s'han desat correctament."));
          }),
          map(usuario => usuarioFormActions.updateUsuarioSuccess({ usuario })),
          catchError(error => of(usuarioFormActions.updateUsuarioFail({ error })))
        );
      })
    )
  );

  deleteUsuario$ = createEffect(() =>
    this.actions$.pipe(
      ofType(usuarioFormActions.deleteUsuario),
      map(action => action.id),
      switchMap(id => {
        return this.usuarioService.delete(id, null).pipe(
          tap(() => {
            this.messageService.sendNotification(
              new BitMessage("Informació", "Les dades s'han eliminat correctament.")
            );
          }),
          map(() => usuarioFormActions.deleteUsuarioSuccess()),
          catchError(error => of(usuarioFormActions.deleteUsuarioFail({ error })))
        );
      })
    )
  );

  insertUsuarioSuccess$ = createEffect(
    () =>
      this.actions$.pipe(
        ofType(usuarioFormActions.insertUsuarioSuccess),
        map(action => action.usuario),
        map(usuario => {
          this.router.navigate(["crud-ngrx8/form", usuario.id], { queryParamsHandling: 'preserve' });
        })
      ),
    { dispatch: false }
  );

  deleteUsuarioSuccess$ = createEffect(
    () =>
      this.actions$.pipe(
        ofType(usuarioFormActions.deleteUsuarioSuccess),
        map(() => {
          this.router.navigate(["crud-ngrx8"], { queryParamsHandling: 'preserve' });
        })
      ),
    { dispatch: false }
  );
}
